#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/10/07 09:06

# Rgb用了numpy，不再属于pupil了，而且会初始化很多颜色列表，也浪费空间，
# 所以改成独立包，需要的时候再手动导入
# from pyxllib.cv.rgbfmt import RgbFormatter



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/11/15 10:16

""" 几何、数学运算

specialist级别
"""

from pyxllib.prog.pupil import check_install_package

check_install_package('cv2', 'opencv-python')

import copy

import numpy as np
import cv2

from pyxllib.algo.intervals import Intervals

____base = """

"""


def xywh2ltrb(p):
    return [p[0], p[1], p[0] + p[2], p[1] + p[3]]


def ltrb2xywh(p):
    return [p[0], p[1], p[2] - p[0], p[3] - p[1]]


def rect2polygon(src_pts):
    """ 矩形对角线两个点，转成四边形四个点的模式来表达
    （输入左上、右下两个顶点坐标）

    :param list|np.ndarray src_pts: size 2*2
    :rtype: list

    >>> rect2polygon([[0, 0], [10, 20]])
    [[0, 0], [10, 0], [10, 20], [0, 20]]
    >>> rect2polygon(np.array([[0, 0], [10, 20]]))
    [[0, 0], [10, 0], [10, 20], [0, 20]]
    >>> rect2polygon([[10, 0], [0, 20]])
    [[0, 0], [10, 0], [10, 20], [0, 20]]
    """
    [[x1, y1], [x2, y2]] = src_pts
    dst_pts = [[x1, y1], [x2, y1], [x2, y2], [x1, y2]]
    dst_pts = resort_quad_points(dst_pts)
    return dst_pts


def reshape_coords(coords, m, dtype=None):
    """ 重置坐标点的维度

    :param list coords: 这个函数主要还是封装了对list情况的处理
        其实np.ndarray结构也行，但这种情况直接用np接口操作就行，不需要引用该函数
    :rtype: list

    # 转成 n*1 的矩阵

    >>> reshape_coords([(1, 2), (3, 4)], 1)
    [1, 2, 3, 4]
    >>> reshape_coords(np.array([[1, 2], [3, 4]]), 1)
    [1, 2, 3, 4]
    >>> reshape_coords([1, 2, 3, 4], 1)
    [1, 2, 3, 4]

    >>> reshape_coords([[1.5, 2], [3.5, 4]], 1)
    [1.5, 2.0, 3.5, 4.0]

    # 这种情况，[3,4]、[5,6,7]都是一个整体
    # VisibleDeprecationWarning
    >>> reshape_coords([1, 2, [3, 4], [5, 6, 7]], 1)
    [1, 2, [3, 4], [5, 6, 7]]

    >>> reshape_coords([[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]], 1)
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]

    # 变成 n*2 的矩阵

    >>> reshape_coords([1, 2, 3, 4], 2)
    [[1, 2], [3, 4]]
    >>> reshape_coords(np.array([1, 2, 3, 4]), 2)
    [[1, 2], [3, 4]]
    >>> reshape_coords([[1, 2], [3, 4]], 2)
    [[1, 2], [3, 4]]
    >>> reshape_coords([1.5, 2, 3.5, 4], 2)
    [[1.5, 2.0], [3.5, 4.0]]
    >>> reshape_coords([1.5, 2, 3.5, 4], 2, dtype=int)  # 数据类型转换
    [[1, 2], [3, 4]]
    """
    if m == 1:
        return np.array(coords, dtype=dtype).reshape(-1).tolist()
    else:
        return np.array(coords, dtype=dtype).reshape((-1, m)).tolist()


def rect_bounds(coords):
    """ 多边形的最大外接矩形

    :param coords: 支持list、np等类型，支持1d、2d两种维度表达方式
    :return: rect的两个点坐标，同时也是 [left, top, right, bottom]
    """
    pts = np.array(coords).reshape(-1).tolist()  # tolist不能删，不然int类型就变了。比如int64不能json.dump
    p = [min(pts[::2]), min(pts[1::2]), max(pts[::2]), max(pts[1::2])]
    return [v for v in p]


def resort_quad_points(src_pts):
    """ 重置四边形点集顺序，确保以左上角为起点，顺时针罗列点集

    算法：先确保pt1、pt2在上面，然后再确保pt1在pt2左边

    :param list|tuple|np.ndarray src_pts: 点集
    :rtype: list|np.ndarray

    >>> pts = [[100, 50], [200, 0], [100, 0], [0, 50]]
    >>> resort_quad_points(pts)
    [[100, 0], [200, 0], [100, 50], [0, 50]]
    >>> pts  # 原来的点不会被修改
    [[100, 50], [200, 0], [100, 0], [0, 50]]

    >>> pts = np.array([[100, 50], [200, 0], [100, 0], [0, 50]])
    >>> resort_quad_points(pts)
    array([[100,   0],
           [200,   0],
           [100,   0],
           [  0,  50]])
    >>> pts  # 原来的点不会被修改
    array([[100,  50],
           [200,   0],
           [100,   0],
           [  0,  50]])
    """
    pts = copy.copy(src_pts)
    if isinstance(pts, np.ndarray):
        pts = pts.tolist()
    if pts[0][1] > pts[2][1]:
        pts[0], pts[2] = pts[2], pts[0]
    if pts[1][1] > pts[3][1]:
        pts[1], pts[3] = pts[3], pts[1]
    if pts[0][0] > pts[1][0]:
        pts[0], pts[1] = pts[1], pts[0]
        pts[2], pts[3] = pts[3], pts[2]
    return pts


def ltrb_border(ltrb, border, size=None):
    """ 给原来的ltrb定位扩展border像素

    Args:
        ltrb:
        border: 可以一个数字，表示统一添加的像素值
            也可以四个数字，表示每个维度分别加的像素值
        size:
            原图的 (width, height)，防止越界
            可以不填，默认不考虑越界问题
    Returns: 新的ltrb坐标
    """
    if isinstance(border, int):
        border = [border] * 4

    l = max(0, ltrb[0] - border[0])
    t = max(0, ltrb[1] - border[1])
    r = min(size[0], ltrb[2] + border[2])
    b = min(size[1], ltrb[3] + border[3])

    return [l, t, r, b]


____warp_perspective = """
仿射、透视变换相关功能

https://www.yuque.com/xlpr/pyxllib/warpperspective
"""


def warp_points(pts, warp_mat):
    """ 透视等点集坐标转换

    :param list|tuple|np.ndarray pts: 支持1d、2d的维度
        其实这个坐标变换就是一个简单的矩阵乘法，只是pts的数据结构往往比较特殊，
        并不是一个n*3的矩阵结构，所以需要进行一些简单的格式转换
        例如 [x1, y1, x2, y2, x3, y3] --> [[x1, x2, x3], [y1, y2, y3], [1, 1, 1]]
    :param list|tuple|np.ndarray warp_mat: 变换矩阵，一般是个3*3的矩阵，但是只输入2*3的矩阵也行，因为第3行并用不到（点集只要取前两个维度X'Y'的结果值）
        TODO 不过这里我有个点也没想明白，如果不用第3行，本质上不是又变回仿射变换了，如何达到透视变换效果？第三维的深度信息能完全舍弃？
    :rtype: np.ndarray

    >>> warp_mat = [[0, 1, 0], [1, 0, 0], [0, 0, 1]]  # 对换x、y
    >>> warp_points([[1, 2], [11, 22]], warp_mat)  # 处理两个点
    array([[ 2,  1],
           [22, 11]])
    >>> warp_points([[1, 2], [11, 22]], [[0, 1, 0], [1, 0, 0]])  # 输入2*3的变换矩阵也可以
    array([[ 2,  1],
           [22, 11]])
    >>> warp_points([1, 2, 11, 22], warp_mat)  # 也可以用一维的结构来输入点集
    array([[ 2,  1],
           [22, 11]])
    >>> warp_points([1, 2, 11, 22, 111, 222], warp_mat)  # 点的数量任意，返回的结构同输入的结构形式
    array([[  2,   1],
           [ 22,  11],
           [222, 111]])
    >>> warp_points(np.array([1, 2, 11, 22, 111, 222]), warp_mat)  # 也可以用np.ndarray等结构
    array([[  2,   1],
           [ 22,  11],
           [222, 111]])
    >>> warp_points([1, 2, 11, 22], warp_mat)  # 也可以用一维的结构来输入点集
    array([[ 2,  1],
           [22, 11]])
    """
    pts1 = np.array(pts).reshape(-1, 2).T
    pts1 = np.concatenate([pts1, [[1] * pts1.shape[1]]], axis=0)
    pts2 = np.dot(warp_mat[:2], pts1)
    pts2 = pts2.T
    return pts2


def get_warp_mat(src, dst):
    """ 从前后点集计算仿射变换矩阵

    :param src: 原点集，支持多种格式输入
    :param dst: 变换后的点集
    :return np.ndarray: 3*3的变换矩阵
    """

    def cvt_data(pts):
        # opencv的透视变换，输入的点集有类型限制，必须使用float32
        return np.array(pts, dtype='float32').reshape((-1, 2))

    src, dst = cvt_data(src), cvt_data(dst)
    n = src.shape[0]
    if n == 3:
        # 只有3个点，则使用仿射变换
        warp_mat = cv2.getAffineTransform(src, dst)
        warp_mat = np.concatenate([warp_mat, [[0, 0, 1]]], axis=0)
    elif n == 4:
        # 有4个点，则使用透视变换
        warp_mat = cv2.getPerspectiveTransform(src, dst)
    else:
        raise ValueError('点集数量过多')
    return warp_mat


def quad_warp_wh(pts, method='average'):
    """ 四边形转为矩形的宽、高

    :param pts: 四个点坐标
        TODO 暂时认为pts是按点集顺时针顺序输入的
        TODO 暂时认为pts[0]就是第一个坐标点
    :param method:
        记四条边分别为w1, h1, w2, h2
        average: 平均宽、高
        max: 最大宽、高
        min: 最小宽、高
    :return: (w, h) 变换后的矩形宽、高
    """
    # 1 计算四边长
    from math import hypot
    # pts = ReshapeCoords.list_2d(pts)
    lens = [0] * 4
    for i in range(4):
        pt1, pt2 = pts[i], pts[(i + 1) % 4]
        lens[i] = hypot(pt1[0] - pt2[0], pt1[1] - pt2[1])

    # 2 目标宽、高
    if method is True:
        method = 'average'
    if method == 'average':
        w, h = (lens[0] + lens[2]) / 2, (lens[1] + lens[3]) / 2
    elif method == 'max':
        w, h = max(lens[0], lens[2]), max(lens[1], lens[3])
    elif method == 'min':
        w, h = min(lens[0], lens[2]), min(lens[1], lens[3])
    else:
        raise ValueError(f'不支持的方法 {method}')
    # 这个主要是用于图像变换的，而图像一般像素坐标要用整数，所以就取整运算了
    return round(w), round(h)


def warp_quad_pts(pts, method='average'):
    """ 将不规则四边形转为矩形

    :param pts: 不规则四边形的四个点坐标
    :param method: 计算矩形宽、高的算法
    :return: 返回时，仍然用四个点的坐标表达，规则矩形的四个点坐标

    >>> warp_quad_pts([[89, 424], [931, 424], [399, 290], [621, 290]])
    [[0, 0], [532, 0], [532, 549], [0, 549]]
    """
    w, h = quad_warp_wh(pts, method)
    return rect2polygon([[0, 0], [w, h]])


____polygon = """
"""


class ComputeIou:
    """ 两个多边形的交并比 Intersection Over Union """

    @classmethod
    def ltrb(cls, pts1, pts2):
        """ https://gist.github.com/meyerjo/dd3533edc97c81258898f60d8978eddc
        """
        # determine the (x, y)-coordinates of the intersection rectangle
        x_a = max(pts1[0], pts2[0])
        y_a = max(pts1[1], pts2[1])
        x_b = min(pts1[2], pts2[2])
        y_b = min(pts1[3], pts2[3])

        # compute the area of intersection rectangle
        inter_area = abs(max((x_b - x_a, 0)) * max((y_b - y_a), 0))
        if inter_area == 0:
            return 0
        # compute the area of both the prediction and ground-truth
        # rectangles
        box_a_area = abs((pts1[2] - pts1[0]) * (pts1[3] - pts1[1]))
        box_b_area = abs((pts2[2] - pts2[0]) * (pts2[3] - pts2[1]))

        # compute the intersection over union by taking the intersection
        # area and dividing it by the sum of prediction + ground-truth
        # areas - the interesection area
        iou = inter_area / float(box_a_area + box_b_area - inter_area)

        # return the intersection over union value
        return iou

    @classmethod
    def polygon(cls, pts1, pts2):
        inter_area = pts1.intersection(pts2).area
        if inter_area:
            union_area = pts1.area + pts2.area - inter_area
            return (inter_area / union_area) if union_area else 0
        else:
            return 0

    @classmethod
    def polygon2(cls, pts1, pts2):
        """ 会强制转为polygon对象再处理

        >>> ComputeIou.polygon2([[0, 0], [10, 10]], [[5, 5], [15, 15]])
        0.14285714285714285
        """
        from pyxllib.algo.shapelylib import ShapelyPolygon
        polygon1, polygon2 = ShapelyPolygon.gen(pts1), ShapelyPolygon.gen(pts2)
        return cls.polygon(polygon1, polygon2)

    @classmethod
    def nms_basic(cls, boxes, func, iou=0.5, *, key=None, index=False):
        """ 假设boxes已经按权重从大到小排过序

        :param boxes: 支持输入一组box列表 [box1, box2, box3, ...]
        :param key: 将框映射为可计算对象
        :param index: 返回不是原始框，而是对应的下标 [i1, i2, i3, ...]
        """
        # 1 映射到items来操作
        if callable(key):
            items = list(enumerate([key(b) for b in boxes]))
        else:
            items = list(enumerate(boxes))

        # 2 正常nms功能
        idxs = []
        while items:
            # 1 加入权值大的框
            i, b = items[0]
            idxs.append(i)
            # 2 抑制其他框
            left_items = []
            for j in range(1, len(items)):
                if func(b, items[j][1]) < iou:
                    left_items.append(items[j])
            items = left_items

        # 3 返回值
        if index:
            return idxs
        else:
            return [boxes[i] for i in idxs]

    @classmethod
    def nms_ltrb(cls, boxes, iou=0.5, *, key=None, index=False):
        return cls.nms_basic(boxes, cls.ltrb, iou, key=key, index=index)

    @classmethod
    def nms_xywh(cls, boxes, iou=0.5, *, key=None, index=False):
        if callable(key):
            func = lambda x: xywh2ltrb(key(x))
        else:
            func = xywh2ltrb
        return cls.nms_ltrb(boxes, iou, key=func, index=index)

    @classmethod
    def nms_polygon(cls, boxes, iou=0.5, *, key=None, index=False):
        # ShapelyPolygon.gen
        return cls.nms_basic(boxes, cls.polygon, iou, key=key, index=index)


____other = """
"""


def divide_quadrangle(coords, r1=0.5, r2=None):
    """ 切分一个四边形为两个四边形

    :param coords: 4*2的坐标
    :param r1: 第一个切分比例，0.5相当于中点（即第一个四边形右边位置）
    :param r2: 第二个切分比例，即第二个四边形左边位置
    :return: 返回切割后所有的四边形

    一般用在改标注结果中，把一个框拆成两个框
    TODO 把接口改成切分一个四边形为任意多个四边形？即把r1、r2等整合为一个list参数输入
    """

    # 1 计算分割点工具
    def segment_point(pt1, pt2, rate=0.5):
        """ 两点间的分割点
        :param rate: 默认0.5是二分点，rate为0时即pt1，rate为1时为pt2，取值可以小于0、大于-1
        :return:
        """
        x1, y1 = pt1
        x2, y2 = pt2
        x, y = x1 + rate * (x2 - x1), y1 + rate * (y2 - y1)
        return int(x), int(y)

    # 2 优化参数值
    # coords = ReshapeCoords.list_2d(coords)
    if not r2: r2 = 1 - r1

    # 3 计算切分后的四边形坐标
    pt1, pt2, pt3, pt4 = coords
    pt5, pt6 = segment_point(pt1, pt2, r1), segment_point(pt4, pt3, r1)
    pt7, pt8 = segment_point(pt1, pt2, r2), segment_point(pt4, pt3, r2)
    return [pt1, pt5, pt6, pt4], [pt7, pt2, pt3, pt8]


def split_vector_interval(vec, maxsplit=None, minwidth=3):
    """
    :param vec: 一个一维向量，需要对这个向量进行切割
        需要前置工作先处理好数值
            使得背景在非正数，背景概率越大，负值绝对值越大
            前景在正值，前景概率越大，数值越大
        要得到能量最大（数值最大、前景内容）的几个区域
        但是因为有噪声的原因，该算法要有一定的抗干扰能力

        一般情况下
            用 0 代表背景
            用 <1 的正值表示这一列黑点所占比例（np.mean）
            用 np.sum 传入整数暂时也行，但考虑以后功能扩展性，用比例会更好
            传入负数，表示特殊背景，该背景可以抵消掉的minwidth宽度数
    :param maxsplit: 最大切分数量，即最多得到几个子区间
        没设置的时候，会对所有满足条件的情况进行切割
    :param minwidth: 每个切分位置最小具有的宽度
    :return: [(l, r), (l, r), ...]  每一段文本的左右区间
    """
    # 1 裁剪左边、右边
    n_vec = len(vec)
    left, right = 0, n_vec
    while left < right and vec[left] <= 0:
        left += 1
    while right > left and vec[right - 1] <= 0:
        right -= 1
    # 左右空白至少也要达到minwidth才去除
    # if left < minwidth: left = 0
    # if n_vec - right + 1 < minwidth: right = n_vec

    vec = vec[left:right]
    width = len(vec)
    if width == 0:
        return []  # 没有内容，返回空list

    # 2 找切分位置
    #   统计每一段连续的背景长度，并且对其数值求和，作为这段是背景的置信度
    bg_probs, bg_start, cnt = [], 0, 0

    def update_fg():
        """ 遇到前景内容，或者循环结束，更新一下 """
        nonlocal cnt
        prob = vec[bg_start:bg_start + cnt].sum()
        # print(cnt, prob)
        if cnt >= (minwidth + prob):  # 负值可以减小minwidth限定
            itv = [bg_start, bg_start + cnt]
            bg_probs.append([itv, prob])
        cnt = 0

    for i in range(width):
        if vec[i] <= 0:
            if not cnt:
                bg_start = i
            cnt += 1
        else:
            update_fg()
    else:
        update_fg()

    # 3 取置信度最大的几个分割点
    if maxsplit:
        bg_probs = sorted(bg_probs, key=lambda x: x[1])[:(maxsplit - 1)]
    bg_probs = sorted(bg_probs, key=lambda x: x[0])  # 从左到右排序

    # 4 返回文本区间（反向计算）
    res = []
    intervals = Intervals([itv for itv, prob in bg_probs]).invert(width) + left
    # print(intervals)
    for interval in intervals:
        res.append([interval.start(), interval.end()])
    return res


def bound_scale(bound, scale):
    """ 一个矩形，以中心为原点，缩放面积为原来scale的新矩形

    :param bound: [x1, y1, x2, y2]
    :param scale: 比例，例如0.5，就是缩放一半
    """
    x1, y1, x2, y2 = bound
    x0 = (x2 + x1) / 2
    y0 = (y2 + y1) / 2
    r = 1 - scale ** 0.5

    x1 += r * abs(x0 - x1)
    y1 += r * abs(y0 - y1)
    x2 -= r * abs(x0 - x2)
    y2 -= r * abs(y0 - y2)
    return x1, y1, x2, y2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/05/19 16:28

""" 相关数据格式类 """

from pathlib import Path
from pyxlpr.ai import *
from pyxlpr.data import *

# from pyxlpr.data.imtextline import TextlineShape

__1_zcdata = """
"""


class ZcTextGt:
    def __init__(self, root, data=None, *, imdir='images', parts=None):
        """
        :param root: 数据根目录
        :param imdir: 所在图片子目录
        :param parts: ['test.txt', 'train.txt'] 等分块标记

        备注：具体文本数据好像不太适合直接存到内存里，就先不存了。
            但是这个类至少可以把相关功能整合在一起，不零散。
        """
        self.root = Dir(root)
        self.imdir = Dir(imdir, self.root)
        self.parts = parts or []

        if data is None:
            pass
        self.data = data

    def writes_from_coco(self, gt_dict, *, prt=False):
        """ coco标注 --> 智财偏好的文本标注格式

        :param gt_dict: coco 的 gt 字典

        TODO gt_dict可能是过量版，增设筛选规则？
        """
        items = list(CocoData(gt_dict).group_gt())
        for img, anns in tqdm(items, desc='生成ZcTextGt的txt标注文件', disable=not prt):
            content = []
            for ann in anns:
                ltrb = xywh2ltrb(ann['bbox'])
                ltrb = ','.join([str(int(v)) for v in ltrb])
                content.append('\t'.join([ltrb, ann['label']]))
            File(img['file_name'], self.imdir, suffix='.txt').write('\n'.join(content))

    def writes(self, *, max_workers=8, prt=False):
        """ 重新写入txt文件 """

        def write(x):
            file, data = x
            if file:  # 如果文件存在，要遵循原有的编码规则
                with open(str(file), 'rb') as f:
                    bstr = f.read()
                encoding = get_encoding(bstr)
                file.write(data, encoding=encoding, if_exists='replace')
            else:  # 否则直接写入
                file.write(data)

        mtqdm(write, self.data, desc='写入labelme json数据', max_workers=max_workers, disable=not prt)


class ZcKvGt(ZcTextGt):

    def writes_from_coco(self, gt_dict, *, prt=False):
        """ coco标注 -> 智财偏好的，带类别的文本标注格式 """
        items = list(CocoData(gt_dict).group_gt())
        for img, anns in tqdm(items, '生成ZcKvGt的txt标注文件', disable=not prt):
            content = []
            for ann in anns:
                ltrb = xywh2ltrb(ann['bbox'])
                ltrb = ','.join([str(int(v)) for v in ltrb])
                cat_id = ann['category_id']
                cat = (cat_id + 1) // 2
                if cat == 5:
                    cat = 0
                    kv = -1
                else:
                    kv = (cat - 1) % 2
                content.append('\t'.join([ltrb, str(cat), str(kv), ann['label']]))
            File(img['file_name'], self.imdir, suffix='.txt').write('\n'.join(content))


class ZcKvDtOld:
    """ 旧版本的解析器 """

    def __init__(self, data=None):
        """
        :param data:
                  {filepath1: [{'logo': True, 'gt': 'address', 'pr': 'address', 'lb': 'LOT 3'},
                              {...},
                              ...]
                  filepath2: ...,
                  ...
                  }
        """
        self.data = data

    @classmethod
    def init_from_zc_txt(cls, file, *, reverse_annos=False):
        """ 从文件解析出字典结构数据

        :return: {filepath1: [{'logo': True, 'gt': 'address', 'pr': 'address', 'lb': 'LOT 3'},
                              {...},
                              ...]
                  filepath2: ...,
                  ...
                  }
        """
        content = File(file).read()
        content = re.sub(r'(.+?)(\n\n)([^\n]+\n)', r'\3\1\2', content, flags=re.DOTALL)
        parts = ContentPartSpliter.multi_blank_lines(content)

        data = dict()
        for pt in parts:
            lines = pt.splitlines()
            filepath = lines[0]
            annos = []
            for line in lines[1:]:  # 第1行是文件名，删掉
                m = re.match(r'(.+?), GT: (.+?), PR: (.+?), LB: (.+)', line)
                logo, gt, pr, lb = m.groups()
                annos.append({'logo': logo == 'True', 'gt': gt, 'pr': pr, 'lb': lb})
            if reverse_annos:
                annos = list(reversed(annos))
            data[filepath] = annos

        return cls(data)

    def to_coco_dt(self, gt_dict, *, printf=True):
        """ 配合gt标注文件，做更精细的zc结果解析
        """
        cat2id = {c['name']: c['id'] for c in gt_dict['categories']}
        gt_annos = {x[0]['file_name']: x for x in CocoData(gt_dict).group_gt()}
        dt_list = []
        for file, dt_annos in self.data.items():
            file_name = pathlib.Path(file).name
            image, annos = gt_annos[file_name]

            # 如果gt的annos比dt_annos少，需要扩充下，默认按最后一个gt的an填充
            n, m = len(annos), len(dt_annos)
            assert n == m  # 智财如果不带box出来，是要强制框数量相同的！否则box怎么一一对应？
            if n < m:
                annos += [annos[-1]] * (m - n)

            # TODO 有些不是按顺序匹配的，增加按文本匹配的功能

            for line, an in zip(dt_annos, annos):
                gt, pr, lb = an['gt'], an['pr'], an['lb']
                if printf and lb != an['label']:
                    print(file_name, lb, '<=>', an['label'])  # gt和dt的框没对应上，最好检查下问题
                # 附加值是识别错误的类别，没有则代表识别正确
                dt_list.append(
                    {'image_id': an['image_id'], 'category_id': cat2id[pr],
                     'bbox': an['bbox'], 'score': 1, 'label': lb})
        return dt_list


class ZcKvDt(ZcKvDtOld):
    """ 智财预测结果文件的通用解析类

    这里是210510周一16:24新版的结果，文件顺序头写对了，并且增加了cs每个结果的置信度
    """

    @classmethod
    def init_from_zc_txt(cls, file, *, reverse_annos=False):
        """ 从文件解析出字典结构数据

        有时候可能没有对应的 coco gt，则可以用这个直接把文件解析为内存数据处理

        :param reverse_annos: 是否对每个图片的标注结果，进行顺序反转
        :return: {filepath1: [{'logo': True, 'gt': 'address', 'pr': 'address', 'lb': 'LOT 3', 'cs': 1.0},
                              {...},
                              ...]
                  filepath2: ...,
                  ...
                  }
        """
        content = File(file).read()
        parts = ContentPartSpliter.multi_blank_lines(content)

        data = dict()
        for pt in parts:
            lines = pt.splitlines()
            filepath = lines[0]
            annos = []
            for line in lines[1:]:  # 第1行是文件名，删掉
                m = re.match(r'(.+?), GT: (.+?), PR: (.+?), LB: (.+?), CS: (.+)', line)
                logo, gt, pr, lb, cs = m.groups()
                annos.append({'logo': logo == 'True', 'gt': gt, 'pr': pr, 'lb': lb, 'cs': float(cs)})
            if reverse_annos:
                annos = list(reversed(annos))
            data[filepath] = annos

        return cls(data)

    def to_coco_dt(self, gt_dict, *, prt=True):
        """ 转coco的dt格式

        :param gt_dict: 需要有参照的gt文件，才能知道图片id，以及补充box位置信息
        """
        cat2id = {c['name']: c['id'] for c in gt_dict['categories']}
        gt_annos = {x[0]['file_name']: x for x in CocoData(gt_dict).group_gt()}
        dt_list = []
        for file, zc_annos in self.data.items():
            file_name = pathlib.Path(file).name
            image, im_annos = gt_annos[file_name]

            # 如果gt的annos比dt_annos少，需要扩充下，默认按最后一个gt的an填充
            n, m = len(im_annos), len(zc_annos)
            if n == m:
                # assert n == m  # 智财如果不带box出来，是要强制框数量相同的！否则box怎么一一对应？
                # if n < m:
                #     annos += [annos[-1]] * (m - n)
                # 要以顺序整体有依据，文本内容匹配为辅的策略配对
                im_annos.sort(key=lambda x: x['label'])
                zc_annos.sort(key=lambda x: x['lb'])
                dt_annos = []

                for a, b in zip(zc_annos, im_annos):
                    gt, pr, lb, cs = a['gt'], a['pr'], a['lb'], a['cs']
                    if prt and lb != b['label']:
                        # gt和dt的框没对应上，最好检查下问题
                        warn = ' '.join([file_name, lb, '<=>', b['label']])
                        dprint(warn)
                    # 附加值是识别错误的类别，没有则代表识别正确
                    dt_annos.append(
                        {'image_id': b['image_id'], 'category_id': cat2id[pr],
                         'bbox': b['bbox'], 'score': cs, 'label': lb})

                # 已经协调确定了空间顺序，但以防万一，可以再按空间排序下给到下游任务
                #   为了效率，也可以确保gt的annos有序，操作时annos顺序不动
                #   这里是前面为了匹配，已经把排序搞乱了，这里是必须要重排
                dt_annos.sort(key=lambda x: TextlineShape(xywh2ltrb(x['bbox'])))  # 几何重排
                dt_list += dt_annos
            else:  # 否则不匹配用不匹配的玩法
                raise NotImplementedError

        return dt_list


__2_other = """
"""


class SroieTextData:
    """ sroie task1、task2 的标注数据

    72,25,326,25,326,64,72,64,TAN WOON YANN
    50,82,440,82,440,121,50,121,BOOK TA .K(TAMAN DAYA) SDN BND
    205,121,285,121,285,139,205,139,789417-W
    """

    def __init__(self, root, part=None):
        """
        :param root: 'data/task1+2'
            images，目录下有987份jpg图片、txt标注
            test.txt，361张jpg图片清单
            train.txt，626张jpg图片清单
        :param part:
            'test', 只记录test的361张图书
            'train' 等同理
        """
        pass

    def to_coco(self):
        pass


class SroieClsData:
    """ sroie task3 关键信息提取的标注 """
    pass


class BaiduOcrRes:
    def __init__(self, data=None):
        """
        :param data: dict (key 文件名 -> value 识别结果dict)
            {'words_result':
                [ {'words': 'xxxx', 'location': {'top': 130, 'left': ~, 'width': ~, 'height': ~}}, ... ]
             'log_id': 136768...
             'words_result_num': 39
            }
        """
        self.data = data

    @classmethod
    def init_from_hxl_txt(cls, file):
        """ 按训龙给的txt格式来初始化

        数据格式：
            一共22499行，大概是所有agreement，但可能有些空白图所以没结果
            每行第一段是png完整路径，第二段是百度api返回的json读取为dict后直接print的结果
        """
        lines = File(file).read().splitlines()
        data = dict()
        for line in tqdm(lines, '解析txt中每张图对应的字典识别数据'):
            if line == '': continue
            # 切分路径和json数据
            # imfile, dictdata = line.split(maxsplit=1)
            imfile, dictdata = re.split(r'\s+', line, maxsplit=1)
            if 'debug' in str(imfile): continue  # 有65张debug_的图片不要（后记：服务器上多余的65个debug图已删）
            data[imfile] = eval(dictdata)

        return cls(data)

    def check(self, imdir):
        """ 检查json数据的一些问题 """
        with TicToc('缺失的文件'):
            files = Path(str(imdir)).glob('*.jpg')
            files1 = {f.stem for f in files}
            files2 = {Path(f).stem for f in self.data.keys()}
            files3 = {Path(f).stem for f in self.data.keys() if ('debug' in str(f))}
            print(f'缺失{len(files1 - files2)}个文件的识别结果')
            print(f'多出{len(files2 - files1)}个文件的识别结果')
            print(f'{len(files3)}个debug文件')
            sys.stderr.flush()

        with TicToc('check errors'):
            ct = Counter()
            for k, v in self.data.items():
                if 'error_code' in v:
                    ct[v['error_code']] += 1
                    print(k, v)
            print(ct.most_common())

    def to_coco_gt(self, images, *, start_dt_id=0):
        """ 转成coco格式

        :param images: coco gt格式的参考images （可以train、val各生成一个返回）
            TODO 可以扩展支持输入图片所在目录的形式初始化的方法
        :param start_dt_id: annotation起始编号
        :return: coco gt格式的字典
        """
        # 辅助数组
        image_files = {Path(x['file_name']).stem: x['id'] for x in images}
        # 遍历判断要处理的文件
        annotations = []
        for k, v in tqdm(self.data.items(), '解析出每张图片识别结果对应的annotations'):
            stem = Path(k).stem
            if stem not in image_files or 'error_code' in v:
                continue
            image_id = image_files[stem]
            for item in v['words_result']:
                loc = item['location']
                bbox = [loc['left'], loc['top'], loc['width'], loc['height']]
                start_dt_id += 1
                an = CocoGtData.gen_annotation(id=start_dt_id, bbox=bbox, image_id=image_id, label=item['words'])
                annotations.append(an)
        return {'images': images,
                'annotations': annotations,
                'categories': CocoGtData.gen_categories(['text'])}


if __name__ == '__main__':
    os.chdir('D:/home/datasets/textGroup/SROIE2019+/data/task3_testcrop')
    with TicToc(__name__):
        ld = LabelmeDataset.init_from_coco(r'test', 'data_crop.json')
        ld.writes()
        # print(ld)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 厦门理工计算机学院 王大寒/(PRIU)福建省模式识别与图像理解重点实验室
# @Document : https://www.yuque.com/xlpr/doai/hesuan
# @Email  : (陈)877362867@qq.com
# @Address: 厦门理工综合楼1905
# @Date   : 2022/08/31

import datetime
import os
import re
import sys
import time

import fire
import numpy as np
import pandas as pd
import requests
from tqdm import tqdm

from PyQt5.QtCore import QSize, Qt, QUrl
from PyQt5.QtGui import QFont, QIcon, QDesktopServices
from PyQt5.QtWidgets import (QApplication, QMainWindow, QLabel, QGridLayout, QWidget,
                             QLineEdit, QToolButton, QFileDialog, QProgressBar)

from pyxllib.file.xlsxlib import openpyxl
from pyxllib.ext.qt import get_input_widget
from pyxllib.xl import TicToc, XlPath, matchpairs, get_etag

from pyxlpr.ai.clientlib import XlAiClient


class PPOCR:
    def __init__(self):
        self.xlapi = XlAiClient()
        self.xlapi.login_priu('hs14TkixP7fD#')
        # 本地缓存，已经识别过的，不重复提交服务器。
        self.cache = self.read_cache()  # etag: result_attrs

    def read_cache(self):
        """ 每个月的结果会在本地做缓存 """
        f = XlPath.tempdir() / datetime.date.today().strftime('hesuan%Y%m.pkl')
        if f.is_file():
            return f.read_pkl()
        else:
            return {}

    def write_cache(self):
        f = XlPath.tempdir() / datetime.date.today().strftime('hesuan%Y%m.pkl')
        f.write_pkl(self.cache)

    def get_image_buffer(self, f):
        # 核酸识别的图片大概率不需要很清晰的图片，每张图文件大小可以控制在300kb以内的jpg文件
        # ratio用不到，api接口只会返回关键类别字典值
        buffer, ratio = self.xlapi.adjust_image(f, limit_b64buffer_size=500 * 1024, b64encode=True)
        return buffer.decode()

    def parse_light(self, f):
        xlapi = self.xlapi
        etag = 'parse_light,' + get_etag(XlPath(f).read_bytes())
        if etag not in self.cache:
            data = {'image': self.get_image_buffer(f)}
            attrs = requests.post(f'{xlapi._priu_host}/api/hesuan/parse_light',
                                  json=data, headers=xlapi._priu_header).json()
            if 'xlapi' in attrs:
                del attrs['xlapi']
            self.cache[etag] = attrs
        return self.cache[etag]

    def parse_multi_layout_light(self, files):
        xlapi = self.xlapi
        etag = 'parse_multi_layout_light,' + ','.join([get_etag(XlPath(f).read_bytes()) for f in files])
        if etag not in self.cache:
            data = {'images': [self.get_image_buffer(x) for x in files]}
            attrs = requests.post(f'{xlapi._priu_host}/api/hesuan/parse_multi_layout_light',
                                  json=data, headers=xlapi._priu_header).json()
            # print(attrs)
            if 'xlapi' in attrs:
                del attrs['xlapi']
            self.cache[etag] = attrs
        return self.cache[etag]


class Hesuan:
    """ 核酸检测的各种功能接口 """

    def __init__(self, task, *, use_online_api=True):
        self.task = task
        # 这里类似配置，把所有需要得变化点封装在这里
        # 双码跟三码用一样的版面分析，但是只保存想要的值，不会保存核酸检测结果之类（无核酸检测报告）
        if self.task == '日常核酸检测报告分析':
            self.columns = ['文件', '姓名', '采样时间', '检测时间', '核酸结果', '联系电话', '身份证号']  # 没有匹配名单，直接生成得列属性
            self.stu_columns = ['班级', '文件', '学号', '姓名', '采样时间', '检测时间', '核酸结果', '身份证号', '联系电话']  # 有匹配名单，直接生成得列属性
        elif self.task == '双码（健康码、行程码）分析':
            self.columns = ['文件', '姓名', '联系电话', '14天经过或途经', '健康码颜色', '身份证号']
            self.stu_columns = ['班级', '文件', '学号', '姓名', '联系电话', '14天经过或途经', '健康码颜色', '身份证号']
            self.file_num = 2
        elif self.task == '双码（健康码、行程码）+ 24小时核酸检测报告分析':
            self.columns = ['文件', '姓名', '采样时间', '检测时间', '核酸结果', '联系电话', '14天经过或途经', '健康码颜色', '身份证号']
            self.stu_columns = ['班级', '文件', '学号', '姓名', '采样时间', '检测时间', '核酸结果',
                                '14天经过或途经', '健康码颜色', '身份证号', '联系电话']
            self.file_num = 3

        self.ppocr = PPOCR() if use_online_api else None

    def ensure_images(self, imdir):
        """ 一些非图片个数数据自动转图片

        其中压缩包只支持zip格式
        """
        # 1 先解压所有的压缩包
        for f in imdir.rglob_files('*.zip'):
            from pyxllib.file.packlib import unpack_archive
            # 没有对应同名目录则解压
            if not f.with_name(f.stem).is_dir():
                unpack_archive(f, wrap=1)

        # 2 再检索所有的pdf
        for f in imdir.rglob_files('*'):
            # 可能存在pdf格式的文件，将其转成jpg图片
            suffix = f.suffix.lower()
            if suffix == '.pdf':
                f2 = f.with_suffix('.jpg')
                if not f2.exists():
                    from pyxllib.file.pdflib import FitzDoc
                    # im = FitzDoc(f).load_page(0).get_pil_image()
                    f = FitzDoc(f)
                    im = f.load_page(0).get_pil_image()
                    im.save(f2)

    def similar(self, x, y):
        """ 计算两条数据间的相似度

        这里不适合用编辑距离，因为levenshtein涉及到c++编译，会让配置变麻烦。
        这里就用正则等一些手段计算相似度。

        相关字段：姓名(+文件)、联系电话、身份证号
        """
        t = 0
        if y['姓名'] in x['文件']:
            t += 100
        if y['姓名'] == x['姓名']:
            t += 200
        elif re.match(re.sub(r'\*+', r'.*', x['姓名']) + '$', y['姓名']):
            t += 50

        def check_key_by_asterisk(k):
            # 带星号*的相似度匹配
            nonlocal t
            if isinstance(y[k], str):
                if y[k] == x[k]:
                    t += 100
                else:
                    if y[k][-2:] == x[k][-2:]:
                        t += 20
                    if y[k][:2] == x[k][:2]:
                        t += 10

        check_key_by_asterisk('联系电话')
        check_key_by_asterisk('身份证号')
        return t

    def link_table(self, df1, df2):
        # 1 找出df1中每一张图片，匹配的是df2中的谁
        idxs = []
        for idx1, row1 in df1.iterrows():
            max_idx2, max_sim = -1, 0
            for idx2, row2 in df2.iterrows():
                sim = self.similar(row1, row2)
                if sim > max_sim:
                    max_idx2 = idx2
                    max_sim = sim
            if max_idx2 == -1:
                # 如果没有匹配到，到时候要直接展示原来的数据
                idxs.append(-idx1)
            else:
                idxs.append(max_idx2)

        for k in ['姓名', '联系电话', '身份证号']:
            df1[k] = [(df2[k][i] if i > 0 else df1[k][-i]) for i in idxs]
        for k in ['班级', '学号']:
            df1[k] = [(df2[k][i] if i > 0 else '') for i in idxs]

        # 2 列出涉及到的班级的成员名单
        classes = set(df1['班级'])  # 0.7版是只显示图片涉及到的班级
        # classes = set(df2['班级'])  # 0.9版是直接以清单中的情况显示
        idxs_ = set(idxs)
        for idx2, row2 in df2.iterrows():
            if idx2 not in idxs_ and row2['班级'] in classes:
                line = {}
                for k in ['文件', '采样时间', '检测时间', '核酸结果']:
                    line[k] = ''
                for k in ['姓名', '联系电话', '身份证号', '班级', '学号']:
                    line[k] = row2[k].strip()
                df1 = df1.append(line, ignore_index=True)

        # 3 重新调整列
        # 保护隐私，不显示电话、身份证
        df = df1[['班级', '文件', '学号', '姓名', '采样时间', '检测时间', '核酸结果']]
        df.sort_values(['班级', '学号'], inplace=True)
        df.index = np.arange(1, len(df) + 1)

        return df

    def link_table2(self, df1, df2):
        # 1 匹配情况
        xs = [row for idx, row in df1.iterrows()]
        ys = [row for idx, row in df2.iterrows()]
        ms = matchpairs(xs, ys, self.similar, least_score=40, index=True)
        idxs = {m[1]: m[0] for m in ms}

        # + 辅助函数
        def quote(s):
            # 带上括号注释
            return f'({s})' if s else ''

        # df2如果有扩展列，也显示出来
        custom_cols = []
        for c in df2.columns:  # 按照原表格的列顺序展示
            c = str(c)
            if c not in {'班级', '学号', '姓名', '身份证号', '联系电话'}:
                custom_cols.append(c)

        def extend_cols(y=None):
            if y is None:
                return [''] * len(custom_cols)
            else:
                return [('' if y[c] != y[c] else y[c]) for c in custom_cols]

        # 2 模板表清单
        ls = []

        # self.stu_colunm, 放在这里方便调试, 分别是日常检测任务的属性列, 双码的任务的属性列, 三码任务的属性列
        # columns = ['班级', '文件', '学号', '姓名', '采样时间', '检测时间', '核酸结果', '身份证号', '联系电话'] + custom_cols
        # colunms = ['班级', '文件', '学号', '姓名', '联系电话', '14天经过或途经', '健康码颜色', '身份证号'] + custom_cols
        # columns = ['班级', '文件', '学号', '姓名', '采样时间', '检测时间', '核酸结果', '14天经过或途经', '健康码颜色', '身份证号', '联系电话'] + custom_cols

        columns = self.stu_columns + custom_cols
        for idx, y in df2.iterrows():
            i = idxs.get(int(idx), -1)
            if self.task == '日常核酸检测报告分析':
                record = [y['班级'], '', y['学号'], y['姓名'], '', '', '', y['身份证号'], y['联系电话']] + extend_cols(y)
            elif self.task == '双码（健康码、行程码）分析':
                record = [y['班级'], '', y['学号'], y['姓名'], y['联系电话'], '', '', y['身份证号']] + extend_cols(y)
            elif self.task == '双码（健康码、行程码）+ 24小时核酸检测报告分析':
                record = [y['班级'], '', y['学号'], y['姓名'], '', '', '', '', '', y['身份证号'], y['联系电话']] + extend_cols(y)
            else:
                raise NotImplementedError

            if i != -1:  # 找得到匹配项
                x = xs[i]
                if self.task == '日常核酸检测报告分析':
                    record[1] = x['文件']
                    record[3] += quote(x['姓名'])
                    record[4] = x['采样时间']
                    record[5] = x['检测时间']
                    record[6] = x['核酸结果']
                    record[7] += quote(x['身份证号'])
                    record[8] += quote(x['联系电话'])
                elif self.task == '双码（健康码、行程码）分析':
                    record[1] = x['文件']
                    record[3] += quote(x['姓名'])
                    record[4] += quote(x['联系电话'])
                    record[5] = x['14天经过或途经']
                    record[6] = x['健康码颜色']
                    record[7] = quote(x['身份证号'])
                elif self.task == '双码（健康码、行程码）+ 24小时核酸检测报告分析':
                    record[1] = x['文件']
                    record[3] += quote(x['姓名'])
                    record[4] = x['采样时间']
                    record[5] = x['检测时间']
                    record[6] = x['核酸结果']
                    record[7] = x['14天经过或途经']
                    record[8] = x['健康码颜色']
                    record[9] += quote(x['身份证号'])
                    record[10] += quote(x['联系电话'])

            ls.append(record)

        # 3 df1中剩余未匹配图片加在最后面
        idxs = {m[0] for m in ms}
        for i, x in enumerate(xs):
            if i not in idxs:
                if self.task == '日常核酸检测报告分析':
                    record = ['', x['文件'], '', quote(x['姓名']),
                              x['采样时间'], x['检测时间'], x['核酸结果'],
                              quote(x['身份证号']), quote(x['联系电话'])]
                elif self.task == '双码（健康码、行程码）分析':
                    record = ['', x['文件'], '', quote(x['姓名']),
                              x['14天经过或途经'], x['健康码颜色'], quote(x['身份证号'])]
                elif self.task == '双码（健康码、行程码）+ 24小时核酸检测报告分析':
                    record = ['', x['文件'], '', quote(x['姓名']),
                              x['采样时间'], x['检测时间'], x['核酸结果'], x['14天经过或途经'], x['健康码颜色'],
                              quote(x['身份证号']), quote(x['联系电话']), ]
                else:
                    raise NotImplementedError

                ls.append(record + extend_cols())

        df = pd.DataFrame.from_records(ls, columns=columns)
        df.index = np.arange(1, len(df) + 1)
        return df

    def parse(self, imdir, students=None, *, pb=None):
        """
        :param imdir: 图片所在目录
        :param students: 【可选】参考学生清单
        :param pb: qt的进度条控件
        """
        # 1 基本的信息解析
        ls = []

        imdir = XlPath(imdir)
        self.ensure_images(imdir)

        # 这里将单码、双码和三码的遍历方式简单区分，双码、三码要求同一个人的两张或者三张放在一个目录下面，单码不要求
        # 单码跟其他两个任务简单区分成两个逻辑
        if self.task == '日常核酸检测报告分析':
            files = list(imdir.glob_images('**/*'))
            total_number = len(files)
            if pb:
                pb.setMaximum(total_number)

            tt = time.time()
            for i, f in tqdm(enumerate(files), '识别中'):
                if pb:
                    pb.setFormat(f'%v/%m，已用时{int(time.time() - tt)}秒')
                    pb.setValue(i)
                    QApplication.processEvents()  # 好像用这个刷新就能避免使用QThread来解决刷新问题
                if f.name == 'xmut.jpg':
                    # 有的人没设目录，直接在根目录全量查找了，此时需要过滤掉我的一个资源图片
                    continue
                try:
                    attrs = self.ppocr.parse_light(f)
                except Exception as e:
                    print(f'{f.as_posix()} 图片有问题，跳过未处理。')
                    continue

                rf, ff = f.relpath(imdir).as_posix(), f.resolve()
                attrs['文件'] = f'<a href="{ff}" target="_blank">{rf}</a><br/>'
                # 如果要展示图片，可以加：<img src="{ff}" width=100/>，但实测效果不好
                row = []
                for col in self.columns:
                    row.append(attrs.get(col, ''))
                ls.append(row)
        else:
            img_lists = self.get_img_fileslist(imdir, self.file_num)
            total_number = len(img_lists)
            if pb:
                pb.setMaximum(total_number)

            tt = time.time()
            for i, img_list in tqdm(enumerate(img_lists), '识别中'):
                if pb:
                    pb.setFormat(f'%v/%m，已用时{int(time.time() - tt)}秒')
                    pb.setValue(i)
                    QApplication.processEvents()  # 好像用这个刷新就能避免使用QThread来解决刷新问题

                if self.file_num == 1 and img_list[0].name == 'xmut.jpg':
                    # 有的人没设目录，直接在根目录全量查找了，此时需要过滤掉我的一个资源图 片
                    continue
                attrs = self.ppocr.parse_multi_layout_light(img_list)

                for f in img_list:
                    rf, ff = f.relpath(imdir).as_posix(), f.resolve()
                    # if attrs['文件'] != '':
                    if '文件' in attrs.keys():
                        attrs['文件'] = attrs['文件'] + ' ' + f'<a href="{ff}" target="_blank">{rf}</a><br/>'
                    else:
                        attrs['文件'] = f'<a href="{ff}" target="_blank">{rf}</a><br/>'
                # 如果要展示图片，可以加：<img src="{ff}" width=100/>，但实测效果不好
                row = []
                for col in self.columns:
                    row.append(attrs.get(col, ''))
                ls.append(row)
        if pb:
            pb.setFormat(f'识别图片%v/%m，已用时{int(time.time() - tt)}秒')
            pb.setValue(total_number)
            QApplication.processEvents()
        df = pd.DataFrame.from_records(ls, columns=self.columns)

        # 2 如果有表格清单，进一步优化数据展示形式
        if students is not None:
            df = self.link_table2(df, students)

        # 3 不同的日期用不同的颜色标记
        # https://www.color-hex.com/
        # 绿，蓝，黄，青，紫，红; 然后深色版本再一套
        colors = ['70ea2a', '187dd8', 'f3cf83', '99ffcc', 'ccaacc', 'ff749e',
                  '138808', '9999ff', 'd19a3f', '99ccaa', 'aa33aa', 'cc0000']

        def set_color(m):
            s1, s2 = m.groups()

            parts = s1.split('-')
            if len(parts) > 1:
                parts[1] = f'{min(12, int(parts[1])):02}'
            s1 = '-'.join(parts)

            # 每个日期颜色是固定的，而不是按照相距今天的天数来选的
            try:
                i = (datetime.date.fromisoformat(s1) - datetime.date(2022, 1, 1)).days
                c = colors[i % len(colors)]
                return f'<td bgcolor="#{c}">{s1}{s2}</td>'
            except ValueError:
                return f'<td>{s1}{s2}</td>'

        res = df.to_html(escape=False)
        res = re.sub(r'<td>(\d{4}-\d{2}-\d{2})(.*?)</td>', set_color, res)
        res = re.sub(r'<td>阳性</td>', r'<td bgcolor="#cc0000">阳性</td>', res)

        return res

    def browser(self):
        """ 打开浏览器查看 """
        pass

    def to_excel(self):
        """ 导出xlsx文件

        我前面功能格式是基于html设计的，这里要导出excel高亮格式的话，还得额外开发工作量。
        而且又会导致一大堆依赖包安装，暂时不考虑扩展。
        """
        raise NotImplementedError

    def get_img_fileslist(self, dir_path, filenum):
        """
        返回目录下大于filenum的文件列表

        :param dirpath:
        :param filenum:
        :return:
        """
        root = XlPath(dir_path)
        files_lists = []
        for _dir in root.rglob_dirs('*'):
            files_list = list(_dir.glob_images('*'))
            if len(files_list) >= filenum:
                files_lists.append(files_list)
        return files_lists


class MainWindow(QMainWindow):
    def __init__(self):
        super(QMainWindow, self).__init__()

        # 设置中文尝试
        menubar = self.menuBar()
        menubar.setNativeMenuBar(False)

        # 1 窗口核心属性
        self.setMinimumSize(QSize(900, 300))
        self.setWindowIcon(QIcon('models/xmut.jpg'))
        self.setWindowTitle("核酸检测分析v4.0（在线版） @(厦门理工学院)福建省模式识别与图像理解重点实验室")

        # 2 使用网格布局
        centralWidget = QWidget(self)
        self.setCentralWidget(centralWidget)

        gridLayout = QGridLayout(self)
        centralWidget.setLayout(gridLayout)

        # 3 每列标签
        for i, text in enumerate(["核酸截图目录：", "人员名单【可选】：", "sheet：", "导出报表：", "检测任务", "进度条："]):
            q = QLabel(text, self)
            q.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
            gridLayout.addWidget(q, i, 0)

        # 4 每行具体功能
        def add_line1():
            """ 第一行的控件，选择截图目录 """
            # 1 输入框
            le = self.srcdirEdit = QLineEdit(self)
            le.setText(os.path.abspath(os.curdir))  # 默认当前目录
            gridLayout.addWidget(le, 0, 1)

            # 2 浏览选择
            def select_srcdir():
                directory = str(QFileDialog.getExistingDirectory()).replace('/', '\\')
                if directory:
                    self.srcdirEdit.setText(directory)

            btn = QToolButton(self)
            btn.setText(u"浏览选择...")
            btn.clicked.connect(select_srcdir)
            gridLayout.addWidget(btn, 0, 2)

        def add_line2():
            """ 第二行的控件，选择excel文件位置 """

            # 1 输入框
            def auto_search_xlsx(root=None):
                """
                :param root: 参考目录，可以不输入，默认以当前工作环境为准
                """
                if root is None:
                    root = XlPath('.')
                root = XlPath(root)

                res = ''
                for f in root.glob('*.xlsx'):
                    res = f  # 取最后一个文件
                return str(res)

            le = self.refxlsxEdit = QLineEdit(self)
            # 自动找当前目录下是否有excel文件，没有则置空
            le.setText(auto_search_xlsx(self.srcdir))
            gridLayout.addWidget(le, 1, 1)

            # 2 浏览选择
            def select_refxlsx():
                file = str(QFileDialog.getOpenFileName(filter='*.xlsx')[0]).replace('/', '\\')
                if file:
                    self.refxlsxEdit.setText(file)

            btn = QToolButton(self)
            btn.setText(u"浏览选择...")
            btn.clicked.connect(select_refxlsx)
            gridLayout.addWidget(btn, 1, 2)

        def add_line3():
            def update_combo_box():
                xlsxfile = self.refxlsx
                if xlsxfile and xlsxfile.is_file():
                    wb = openpyxl.open(str(xlsxfile), read_only=True, data_only=True)
                    self.sheets_cb.reset_items(tuple([ws.title for ws in wb.worksheets]))
                else:
                    self.sheets_cb.reset_items(tuple())

            self.sheets_cb = get_input_widget(tuple(), parent=self)
            gridLayout.addWidget(self.sheets_cb, 2, 1)
            update_combo_box()
            self.refxlsxEdit.textChanged.connect(update_combo_box)

        def add_line4():
            self.dstfileEdit = QLineEdit(self)
            self.dstfileEdit.setText(str(XlPath('./报表.html').absolute()))
            gridLayout.addWidget(self.dstfileEdit, 3, 1)

        def add_line5():
            # 任务选择、单码、双码、三码、
            def get_value():
                self.task = self.task_cb.currentText()

            self.task_cb = get_input_widget(('日常核酸检测报告分析',
                                             '双码（健康码、行程码）分析',
                                             '双码（健康码、行程码）+ 24小时核酸检测报告分析'),
                                            cur_value='日常核酸检测报告分析', parent=self)
            self.task = '日常核酸检测报告分析'
            gridLayout.addWidget(self.task_cb, 4, 1)
            self.task_cb.currentTextChanged.connect(get_value)

        def add_line6():
            # 进度条
            pb = self.pb = QProgressBar(self)
            pb.setAlignment(Qt.AlignHCenter)
            gridLayout.addWidget(pb, 5, 1)

        def add_line7():
            # 1 生成报表
            btn = QToolButton(self)
            btn.setText("生成报表")
            btn.setFont(QFont('Times', 20))
            btn.clicked.connect(self.stat)
            gridLayout.addWidget(btn, 6, 1)

            # 2 帮助文档
            def help():
                url = QUrl("https://www.yuque.com/xlpr/doai/hesuan")
                QDesktopServices.openUrl(url)

            btn = QToolButton(self)
            btn.setText("使用手册")
            btn.setFont(QFont('Times', 20))
            btn.clicked.connect(help)
            gridLayout.addWidget(btn, 6, 1, alignment=Qt.AlignHCenter)

        add_line1()
        add_line2()
        add_line3()
        add_line4()
        add_line5()
        add_line6()
        add_line7()

    @property
    def srcdir(self):
        return XlPath(self.srcdirEdit.text())

    @property
    def refxlsx(self):
        p = XlPath(self.refxlsxEdit.text())
        if p.is_file():
            return p

    def get_sheet_name(self):
        return self.sheets_cb.currentText()

    def get_students(self):
        if self.refxlsx:
            df = pd.read_excel(self.refxlsx, self.get_sheet_name(),
                               dtype={'学号': str, '联系电话': str, '身份证号': str, '姓名': str})
            df = df[(~df['姓名'].isna()) & (~df['班级'].isna()) & (~df['学号'].isna())]
            return df
        return None

    @property
    def dstfile(self):
        return XlPath(self.dstfileEdit.text())

    def stat(self):
        hs = Hesuan(task=self.task)
        res = hs.parse(self.srcdir, self.get_students(), pb=self.pb)
        dstfile = self.dstfile
        dstfile.write_text(res, encoding='utf8')
        hs.ppocr.write_cache()
        os.startfile(dstfile)


def gui():
    """ 打开可视化窗口使用 """
    app = QApplication(sys.argv)
    mainWin = MainWindow()
    mainWin.show()
    sys.exit(app.exec_())


def 性能测试():
    model = PPOCR()

    files = list(XlPath('/home/chenkunze/data/hesuan/data/A1 核酸报告-按模板分类').rglob_images())
    tt = time.time()
    for f in tqdm(files):
        model.parse_light(f)
    print((time.time() - tt) / len(files), '秒/张')

    model.write_cache()


if __name__ == '__main__':
    if len(sys.argv) == 1:  # 默认执行main函数
        sys.argv += ['gui']
    with TicToc(__name__):
        fire.Fire()



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ClsMetric(object):
    def __init__(self, main_indicator='acc', **kwargs):
        self.main_indicator = main_indicator
        self.reset()

    def __call__(self, pred_label, *args, **kwargs):
        preds, labels = pred_label
        correct_num = 0
        all_num = 0
        for (pred, pred_conf), (target, _) in zip(preds, labels):
            if pred == target:
                correct_num += 1
            all_num += 1
        self.correct_num += correct_num
        self.all_num += all_num
        return {'acc': correct_num / all_num, }

    def get_metric(self):
        """
        return metrics {
                 'acc': 0
            }
        """
        acc = self.correct_num / self.all_num
        self.reset()
        return {'acc': acc}

    def reset(self):
        self.correct_num = 0
        self.all_num = 0



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '..')))

from pyxlpr.ppocr.utils.logging import get_logger
logger = get_logger()

import cv2
import numpy as np
import time
from PIL import Image
from pyxlpr.ppocr.utils.utility import get_image_file_list
from pyxlpr.ppocr.tools.infer.utility import draw_ocr, draw_boxes

import requests
import json
import base64


def cv2_to_base64(image):
    return base64.b64encode(image).decode('utf8')


def draw_server_result(image_file, res):
    img = cv2.imread(image_file)
    image = Image.fromarray(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
    if len(res) == 0:
        return np.array(image)
    keys = res[0].keys()
    if 'text_region' not in keys:  # for ocr_rec, draw function is invalid 
        logger.info("draw function is invalid for ocr_rec!")
        return None
    elif 'text' not in keys:  # for ocr_det
        logger.info("draw text boxes only!")
        boxes = []
        for dno in range(len(res)):
            boxes.append(res[dno]['text_region'])
        boxes = np.array(boxes)
        draw_img = draw_boxes(image, boxes)
        return draw_img
    else:  # for ocr_system
        logger.info("draw boxes and texts!")
        boxes = []
        texts = []
        scores = []
        for dno in range(len(res)):
            boxes.append(res[dno]['text_region'])
            texts.append(res[dno]['text'])
            scores.append(res[dno]['confidence'])
        boxes = np.array(boxes)
        scores = np.array(scores)
        draw_img = draw_ocr(
            image, boxes, texts, scores, draw_txt=True, drop_score=0.5)
        return draw_img


def main(url, image_path):
    image_file_list = get_image_file_list(image_path)
    is_visualize = False
    headers = {"Content-type": "application/json"}
    cnt = 0
    total_time = 0
    for image_file in image_file_list:
        img = open(image_file, 'rb').read()
        if img is None:
            logger.info("error in loading image:{}".format(image_file))
            continue

        # 发送HTTP请求
        starttime = time.time()
        data = {'images': [cv2_to_base64(img)]}
        r = requests.post(url=url, headers=headers, data=json.dumps(data))
        elapse = time.time() - starttime
        total_time += elapse
        logger.info("Predict time of %s: %.3fs" % (image_file, elapse))
        res = r.json()["results"][0]
        logger.info(res)

        if is_visualize:
            draw_img = draw_server_result(image_file, res)
            if draw_img is not None:
                draw_img_save = "./server_results/"
                if not os.path.exists(draw_img_save):
                    os.makedirs(draw_img_save)
                cv2.imwrite(
                    os.path.join(draw_img_save, os.path.basename(image_file)),
                    draw_img[:, :, ::-1])
                logger.info("The visualized image saved in {}".format(
                    os.path.join(draw_img_save, os.path.basename(image_file))))
        cnt += 1
        if cnt % 100 == 0:
            logger.info("{} processed".format(cnt))
    logger.info("avg time cost: {}".format(float(total_time) / cnt))


if __name__ == '__main__':
    if len(sys.argv) != 3:
        logger.info("Usage: %s server_url image_path" % sys.argv[0])
    else:
        server_url = sys.argv[1]
        image_path = sys.argv[2]
        main(server_url, image_path)



# copyright (c) 2019 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import paddle
from paddle import nn
import paddle.nn.functional as F
from paddle import ParamAttr


class ConvBNLayer(nn.Layer):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride,
                 padding,
                 groups=1,
                 if_act=True,
                 act=None,
                 name=None):
        super(ConvBNLayer, self).__init__()
        self.if_act = if_act
        self.act = act
        self.conv = nn.Conv2D(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=groups,
            weight_attr=ParamAttr(name=name + '_weights'),
            bias_attr=False)

        self.bn = nn.BatchNorm(
            num_channels=out_channels,
            act=act,
            param_attr=ParamAttr(name="bn_" + name + "_scale"),
            bias_attr=ParamAttr(name="bn_" + name + "_offset"),
            moving_mean_name="bn_" + name + "_mean",
            moving_variance_name="bn_" + name + "_variance")

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        return x


class EASTHead(nn.Layer):
    """
    """
    def __init__(self, in_channels, model_name, **kwargs):
        super(EASTHead, self).__init__()
        self.model_name = model_name
        if self.model_name == "large":
            num_outputs = [128, 64, 1, 8]
        else:
            num_outputs = [64, 32, 1, 8]

        self.det_conv1 = ConvBNLayer(
            in_channels=in_channels,
            out_channels=num_outputs[0],
            kernel_size=3,
            stride=1,
            padding=1,
            if_act=True,
            act='relu',
            name="det_head1")
        self.det_conv2 = ConvBNLayer(
            in_channels=num_outputs[0],
            out_channels=num_outputs[1],
            kernel_size=3,
            stride=1,
            padding=1,
            if_act=True,
            act='relu',
            name="det_head2")
        self.score_conv = ConvBNLayer(
            in_channels=num_outputs[1],
            out_channels=num_outputs[2],
            kernel_size=1,
            stride=1,
            padding=0,
            if_act=False,
            act=None,
            name="f_score")
        self.geo_conv = ConvBNLayer(
            in_channels=num_outputs[1],
            out_channels=num_outputs[3],
            kernel_size=1,
            stride=1,
            padding=0,
            if_act=False,
            act=None,
            name="f_geo")

    def forward(self, x, targets=None):
        f_det = self.det_conv1(x)
        f_det = self.det_conv2(f_det)
        f_score = self.score_conv(f_det)
        f_score = F.sigmoid(f_score)
        f_geo = self.geo_conv(f_det)
        f_geo = (F.sigmoid(f_geo) - 0.5) * 2 * 800

        pred = {'f_score': f_score, 'f_geo': f_geo}
        return pred



# copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import os
from paddle.io import Dataset
from .imaug import transform, create_operators
import random


class PGDataSet(Dataset):
    def __init__(self, config, mode, logger, seed=None):
        super(PGDataSet, self).__init__()

        self.logger = logger
        self.seed = seed
        self.mode = mode
        global_config = config['Global']
        dataset_config = config[mode]['dataset']
        loader_config = config[mode]['loader']

        self.delimiter = dataset_config.get('delimiter', '\t')
        label_file_list = dataset_config.pop('label_file_list')
        data_source_num = len(label_file_list)
        ratio_list = dataset_config.get("ratio_list", [1.0])
        if isinstance(ratio_list, (float, int)):
            ratio_list = [float(ratio_list)] * int(data_source_num)
        assert len(
            ratio_list
        ) == data_source_num, "The length of ratio_list should be the same as the file_list."
        self.data_dir = dataset_config['data_dir']
        self.do_shuffle = loader_config['shuffle']

        logger.info("Initialize indexs of datasets:%s" % label_file_list)
        self.data_lines = self.get_image_info_list(label_file_list, ratio_list)
        self.data_idx_order_list = list(range(len(self.data_lines)))
        if mode.lower() == "train":
            self.shuffle_data_random()

        self.ops = create_operators(dataset_config['transforms'], global_config)

    def shuffle_data_random(self):
        if self.do_shuffle:
            random.seed(self.seed)
            random.shuffle(self.data_lines)
        return

    def get_image_info_list(self, file_list, ratio_list):
        if isinstance(file_list, str):
            file_list = [file_list]
        data_lines = []
        for idx, file in enumerate(file_list):
            with open(file, "rb") as f:
                lines = f.readlines()
                if self.mode == "train" or ratio_list[idx] < 1.0:
                    random.seed(self.seed)
                    lines = random.sample(lines,
                                          round(len(lines) * ratio_list[idx]))
                data_lines.extend(lines)
        return data_lines

    def __getitem__(self, idx):
        file_idx = self.data_idx_order_list[idx]
        data_line = self.data_lines[file_idx]
        img_id = 0
        try:
            data_line = data_line.decode('utf-8')
            substr = data_line.strip("\n").split(self.delimiter)
            file_name = substr[0]
            label = substr[1]
            img_path = os.path.join(self.data_dir, file_name)
            if self.mode.lower() == 'eval':
                try:
                    img_id = int(data_line.split(".")[0][7:])
                except:
                    img_id = 0
            data = {'img_path': img_path, 'label': label, 'img_id': img_id}
            if not os.path.exists(img_path):
                raise Exception("{} does not exist!".format(img_path))
            with open(data['img_path'], 'rb') as f:
                img = f.read()
                data['image'] = img
            outs = transform(data, self.ops)
        except Exception as e:
            self.logger.error(
                "When parsing line {}, error happened with msg: {}".format(
                    self.data_idx_order_list[idx], e))
            outs = None
        if outs is None:
            return self.__getitem__(np.random.randint(self.__len__()))
        return outs

    def __len__(self):
        return len(self.data_idx_order_list)
